@@ -3,6 +3,7 @@
 // For conditions of distribution and use, see copyright notice in irrlicht.h
 
 #include <algorithm>
+#include <cassert>
 
 #include "CSceneManager.h"
 #include "IVideoDriver.h"
@@ -41,7 +42,7 @@ CSceneManager::CSceneManager(video::IVideoDriver *driver,
 		ISceneNode(0, 0),
 		Driver(driver),
 		CursorControl(cursorControl),
-		ActiveCamera(0), Parameters(0),
+		ActiveCamera(0),
 		MeshCache(cache), CurrentRenderPass(ESNRP_NONE)
 {
 	// root node's scene manager
@@ -59,9 +60,6 @@ CSceneManager::CSceneManager(video::IVideoDriver *driver,
 	else
 		MeshCache->grab();
 
-	// set scene parameters
-	Parameters = new io::CAttributes();
-
 	// create collision manager
 	CollisionManager = new CSceneCollisionManager(this, Driver);
 
@@ -104,9 +102,6 @@ CSceneManager::~CSceneManager()
 	if (MeshCache)
 		MeshCache->drop();
 
-	if (Parameters)
-		Parameters->drop();
-
 	// remove all nodes before dropping the driver
 	// as render targets may be destroyed twice
 
@@ -305,7 +300,7 @@ void CSceneManager::render()
 //! returns the axis aligned bounding box of this node
 const core::aabbox3d<f32> &CSceneManager::getBoundingBox() const
 {
-	_IRR_DEBUG_BREAK_IF(true) // Bounding Box of Scene Manager should never be used.
+	assert(false); // Bounding Box of Scene Manager should never be used.
 
 	static const core::aabbox3d<f32> dummy{{0.0f, 0.0f, 0.0f}};
 	return dummy;
@@ -471,8 +466,7 @@ void CSceneManager::drawAll()
 	Driver->setTransform(video::ETS_WORLD, core::IdentityMatrix);
 	for (u32 i = video::ETS_COUNT - 1; i >= video::ETS_TEXTURE_0; --i)
 		Driver->setTransform((video::E_TRANSFORMATION_STATE)i, core::IdentityMatrix);
-	// TODO: This should not use an attribute here but a real parameter when necessary (too slow!)
-	Driver->setAllowZWriteOnTransparent(Parameters->getAttributeAsBool(ALLOW_ZWRITE_ON_TRANSPARENT));
+	Driver->setAllowZWriteOnTransparent(true);
 
 	// do animations and other stuff.
 	OnAnimate(os::Timer::getTime());
@@ -742,12 +736,6 @@ void CSceneManager::clear()
 	removeAll();
 }
 
-//! Returns interface to the parameters set in this scene.
-io::IAttributes *CSceneManager::getParameters()
-{
-	return Parameters;
-}
-
 //! Returns current render pass.
 E_SCENE_NODE_RENDER_PASS CSceneManager::getSceneNodeRenderPass() const
 {
@@ -774,7 +762,7 @@ ISceneManager *CSceneManager::createNewSceneManager(bool cloneContent)
 //! Get a skinned mesh, which is not available as header-only code
 SkinnedMesh *CSceneManager::createSkinnedMesh()
 {
-	return new SkinnedMesh();
+	return new SkinnedMesh(SkinnedMesh::SourceFormat::OTHER);
 }
 
 // creates a scenemanager
