@@ -292,12 +292,13 @@ void Map::timerUpdate(float dtime, float unload_timeout, s32 max_loaded_blocks,
 
 	// If there is no practical limit, we spare creation of mapblock_queue
 	if (max_loaded_blocks < 0) {
+		MapBlockVect blocks;
 		for (auto &sector_it : m_sectors) {
 			MapSector *sector = sector_it.second;
 
 			bool all_blocks_deleted = true;
 
-			MapBlockVect blocks;
+			blocks.clear();
 			sector->getBlocks(blocks);
 
 			for (MapBlock *block : blocks) {
@@ -336,10 +337,11 @@ void Map::timerUpdate(float dtime, float unload_timeout, s32 max_loaded_blocks,
 		}
 	} else {
 		std::priority_queue<TimeOrderedMapBlock> mapblock_queue;
+		MapBlockVect blocks;
 		for (auto &sector_it : m_sectors) {
 			MapSector *sector = sector_it.second;
 
-			MapBlockVect blocks;
+			blocks.clear();
 			sector->getBlocks(blocks);
 
 			for (MapBlock *block : blocks) {
@@ -417,16 +419,16 @@ void Map::timerUpdate(float dtime, float unload_timeout, s32 max_loaded_blocks,
 
 void Map::unloadUnreferencedBlocks(std::vector<v3s16> *unloaded_blocks)
 {
-	timerUpdate(0.0, -1.0, 0, unloaded_blocks);
+	timerUpdate(0, -1, 0, unloaded_blocks);
 }
 
-void Map::deleteSectors(std::vector<v2s16> &sectorList)
+void Map::deleteSectors(const std::vector<v2s16> &sectorList)
 {
 	for (v2s16 j : sectorList) {
 		MapSector *sector = m_sectors[j];
 		// If sector is in sector cache, remove it from there
-		if(m_sector_cache == sector)
-			m_sector_cache = NULL;
+		if (m_sector_cache == sector)
+			m_sector_cache = nullptr;
 		// Remove from map and delete
 		m_sectors.erase(j);
 		delete sector;
